# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: indexify/proto/executor_api.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC, 6, 31, 1, "", "indexify/proto/executor_api.proto"
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n!indexify/proto/executor_api.proto\x12\x0f\x65xecutor_api_pb"\xbf\x03\n\x0b\x44\x61taPayload\x12\x10\n\x03uri\x18\x01 \x01(\tH\x00\x88\x01\x01\x12;\n\x08\x65ncoding\x18\x02 \x01(\x0e\x32$.executor_api_pb.DataPayloadEncodingH\x01\x88\x01\x01\x12\x1d\n\x10\x65ncoding_version\x18\x03 \x01(\x04H\x02\x88\x01\x01\x12\x19\n\x0c\x63ontent_type\x18\x04 \x01(\tH\x03\x88\x01\x01\x12\x1a\n\rmetadata_size\x18\x05 \x01(\x04H\x04\x88\x01\x01\x12\x13\n\x06offset\x18\x06 \x01(\x04H\x05\x88\x01\x01\x12\x11\n\x04size\x18\x07 \x01(\x04H\x06\x88\x01\x01\x12\x18\n\x0bsha256_hash\x18\x08 \x01(\tH\x07\x88\x01\x01\x12$\n\x17source_function_call_id\x18\t \x01(\tH\x08\x88\x01\x01\x12\x0f\n\x02id\x18\n \x01(\tH\t\x88\x01\x01\x42\x06\n\x04_uriB\x0b\n\t_encodingB\x13\n\x11_encoding_versionB\x0f\n\r_content_typeB\x10\n\x0e_metadata_sizeB\t\n\x07_offsetB\x07\n\x05_sizeB\x0e\n\x0c_sha256_hashB\x1a\n\x18_source_function_call_idB\x05\n\x03_id"e\n\x0cGPUResources\x12\x12\n\x05\x63ount\x18\x01 \x01(\rH\x00\x88\x01\x01\x12-\n\x05model\x18\x02 \x01(\x0e\x32\x19.executor_api_pb.GPUModelH\x01\x88\x01\x01\x42\x08\n\x06_countB\x08\n\x06_model"\xc2\x01\n\rHostResources\x12\x16\n\tcpu_count\x18\x01 \x01(\rH\x00\x88\x01\x01\x12\x19\n\x0cmemory_bytes\x18\x02 \x01(\x04H\x01\x88\x01\x01\x12\x17\n\ndisk_bytes\x18\x03 \x01(\x04H\x02\x88\x01\x01\x12/\n\x03gpu\x18\x04 \x01(\x0b\x32\x1d.executor_api_pb.GPUResourcesH\x03\x88\x01\x01\x42\x0c\n\n_cpu_countB\x0f\n\r_memory_bytesB\r\n\x0b_disk_bytesB\x06\n\x04_gpu"\xd3\x01\n\x0f\x41llowedFunction\x12\x16\n\tnamespace\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x1d\n\x10\x61pplication_name\x18\x02 \x01(\tH\x01\x88\x01\x01\x12\x1a\n\rfunction_name\x18\x03 \x01(\tH\x02\x88\x01\x01\x12 \n\x13\x61pplication_version\x18\x04 \x01(\tH\x03\x88\x01\x01\x42\x0c\n\n_namespaceB\x13\n\x11_application_nameB\x10\n\x0e_function_nameB\x16\n\x14_application_version"\xd8\x01\n\x19\x46unctionExecutorResources\x12\x1b\n\x0e\x63pu_ms_per_sec\x18\x01 \x01(\rH\x00\x88\x01\x01\x12\x19\n\x0cmemory_bytes\x18\x02 \x01(\x04H\x01\x88\x01\x01\x12\x17\n\ndisk_bytes\x18\x03 \x01(\x04H\x02\x88\x01\x01\x12/\n\x03gpu\x18\x04 \x01(\x0b\x32\x1d.executor_api_pb.GPUResourcesH\x03\x88\x01\x01\x42\x11\n\x0f_cpu_ms_per_secB\x0f\n\r_memory_bytesB\r\n\x0b_disk_bytesB\x06\n\x04_gpu"\xcf\x01\n\x0b\x46unctionRef\x12\x16\n\tnamespace\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x1d\n\x10\x61pplication_name\x18\x02 \x01(\tH\x01\x88\x01\x01\x12\x1a\n\rfunction_name\x18\x03 \x01(\tH\x02\x88\x01\x01\x12 \n\x13\x61pplication_version\x18\x04 \x01(\tH\x03\x88\x01\x01\x42\x0c\n\n_namespaceB\x13\n\x11_application_nameB\x10\n\x0e_function_nameB\x16\n\x14_application_version"\xa3\x04\n\x1b\x46unctionExecutorDescription\x12\x0f\n\x02id\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x33\n\x08\x66unction\x18\x02 \x01(\x0b\x32\x1c.executor_api_pb.FunctionRefH\x01\x88\x01\x01\x12\x14\n\x0csecret_names\x18\x03 \x03(\t\x12&\n\x19initialization_timeout_ms\x18\x04 \x01(\rH\x02\x88\x01\x01\x12\x36\n\x0b\x61pplication\x18\x05 \x01(\x0b\x32\x1c.executor_api_pb.DataPayloadH\x03\x88\x01\x01\x12\x42\n\tresources\x18\x06 \x01(\x0b\x32*.executor_api_pb.FunctionExecutorResourcesH\x04\x88\x01\x01\x12&\n\x19output_payload_uri_prefix\x18\x07 \x01(\tH\x05\x88\x01\x01\x12\x1c\n\x0fmax_concurrency\x18\x08 \x01(\rH\x06\x88\x01\x01\x12"\n\x15\x61llocation_timeout_ms\x18\t \x01(\rH\x07\x88\x01\x01\x42\x05\n\x03_idB\x0b\n\t_functionB\x1c\n\x1a_initialization_timeout_msB\x0e\n\x0c_applicationB\x0c\n\n_resourcesB\x1c\n\x1a_output_payload_uri_prefixB\x12\n\x10_max_concurrencyB\x18\n\x16_allocation_timeout_ms"\xcf\x02\n\x15\x46unctionExecutorState\x12\x46\n\x0b\x64\x65scription\x18\x01 \x01(\x0b\x32,.executor_api_pb.FunctionExecutorDescriptionH\x00\x88\x01\x01\x12<\n\x06status\x18\x02 \x01(\x0e\x32\'.executor_api_pb.FunctionExecutorStatusH\x01\x88\x01\x01\x12S\n\x12termination_reason\x18\x03 \x01(\x0e\x32\x32.executor_api_pb.FunctionExecutorTerminationReasonH\x02\x88\x01\x01\x12)\n!allocation_ids_caused_termination\x18\x04 \x03(\tB\x0e\n\x0c_descriptionB\t\n\x07_statusB\x15\n\x13_termination_reason"\xce\x05\n\rExecutorState\x12\x18\n\x0b\x65xecutor_id\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x15\n\x08hostname\x18\x03 \x01(\tH\x01\x88\x01\x01\x12\x14\n\x07version\x18\x05 \x01(\tH\x02\x88\x01\x01\x12\x34\n\x06status\x18\x06 \x01(\x0e\x32\x1f.executor_api_pb.ExecutorStatusH\x03\x88\x01\x01\x12<\n\x0ftotal_resources\x18\r \x01(\x0b\x32\x1e.executor_api_pb.HostResourcesH\x04\x88\x01\x01\x12N\n!total_function_executor_resources\x18\x07 \x01(\x0b\x32\x1e.executor_api_pb.HostResourcesH\x05\x88\x01\x01\x12;\n\x11\x61llowed_functions\x18\x08 \x03(\x0b\x32 .executor_api_pb.AllowedFunction\x12H\n\x18\x66unction_executor_states\x18\t \x03(\x0b\x32&.executor_api_pb.FunctionExecutorState\x12:\n\x06labels\x18\n \x03(\x0b\x32*.executor_api_pb.ExecutorState.LabelsEntry\x12\x17\n\nstate_hash\x18\x0b \x01(\tH\x06\x88\x01\x01\x12\x19\n\x0cserver_clock\x18\x0c \x01(\x04H\x07\x88\x01\x01\x1a-\n\x0bLabelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x0e\n\x0c_executor_idB\x0b\n\t_hostnameB\n\n\x08_versionB\t\n\x07_statusB\x12\n\x10_total_resourcesB$\n"_total_function_executor_resourcesB\r\n\x0b_state_hashB\x0f\n\r_server_clock"y\n\x0e\x45xecutorUpdate\x12\x18\n\x0b\x65xecutor_id\x18\x01 \x01(\tH\x00\x88\x01\x01\x12=\n\x12\x61llocation_results\x18\x02 \x03(\x0b\x32!.executor_api_pb.AllocationResultB\x0e\n\x0c_executor_id"\xbf\x01\n\x1aReportExecutorStateRequest\x12;\n\x0e\x65xecutor_state\x18\x01 \x01(\x0b\x32\x1e.executor_api_pb.ExecutorStateH\x00\x88\x01\x01\x12=\n\x0f\x65xecutor_update\x18\x02 \x01(\x0b\x32\x1f.executor_api_pb.ExecutorUpdateH\x01\x88\x01\x01\x42\x11\n\x0f_executor_stateB\x12\n\x10_executor_update"\x1d\n\x1bReportExecutorStateResponse"\x9a\x04\n\nAllocation\x12\x33\n\x08\x66unction\x18\x01 \x01(\x0b\x32\x1c.executor_api_pb.FunctionRefH\x00\x88\x01\x01\x12\x1a\n\rallocation_id\x18\x02 \x01(\tH\x01\x88\x01\x01\x12\x1d\n\x10\x66unction_call_id\x18\x03 \x01(\tH\x02\x88\x01\x01\x12\x17\n\nrequest_id\x18\x04 \x01(\tH\x03\x88\x01\x01\x12*\n\x04\x61rgs\x18\x05 \x03(\x0b\x32\x1c.executor_api_pb.DataPayload\x12&\n\x19output_payload_uri_prefix\x18\x06 \x01(\tH\x04\x88\x01\x01\x12-\n request_error_payload_uri_prefix\x18\x07 \x01(\tH\x05\x88\x01\x01\x12!\n\x14\x66unction_executor_id\x18\x08 \x01(\tH\x06\x88\x01\x01\x12#\n\x16\x66unction_call_metadata\x18\t \x01(\x0cH\x07\x88\x01\x01\x42\x0b\n\t_functionB\x10\n\x0e_allocation_idB\x13\n\x11_function_call_idB\r\n\x0b_request_idB\x1c\n\x1a_output_payload_uri_prefixB#\n!_request_error_payload_uri_prefixB\x17\n\x15_function_executor_idB\x19\n\x17_function_call_metadata"K\n\x1fGetDesiredExecutorStatesRequest\x12\x18\n\x0b\x65xecutor_id\x18\x01 \x01(\tH\x00\x88\x01\x01\x42\x0e\n\x0c_executor_id"\xb0\x01\n\x14\x44\x65siredExecutorState\x12H\n\x12\x66unction_executors\x18\x01 \x03(\x0b\x32,.executor_api_pb.FunctionExecutorDescription\x12\x30\n\x0b\x61llocations\x18\x02 \x03(\x0b\x32\x1b.executor_api_pb.Allocation\x12\x12\n\x05\x63lock\x18\x03 \x01(\x04H\x00\x88\x01\x01\x42\x08\n\x06_clock"\x80\x01\n\x13\x45xecutionPlanUpdate\x12\x36\n\rfunction_call\x18\n \x01(\x0b\x32\x1d.executor_api_pb.FunctionCallH\x00\x12+\n\x06reduce\x18\x0c \x01(\x0b\x32\x19.executor_api_pb.ReduceOpH\x00\x42\x04\n\x02op"\xbe\x01\n\x0c\x46unctionCall\x12\x0f\n\x02id\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x31\n\x06target\x18\x02 \x01(\x0b\x32\x1c.executor_api_pb.FunctionRefH\x01\x88\x01\x01\x12*\n\x04\x61rgs\x18\x03 \x03(\x0b\x32\x1c.executor_api_pb.FunctionArg\x12\x1a\n\rcall_metadata\x18\x05 \x01(\x0cH\x02\x88\x01\x01\x42\x05\n\x03_idB\t\n\x07_targetB\x10\n\x0e_call_metadata"h\n\x0b\x46unctionArg\x12\x1a\n\x10\x66unction_call_id\x18\x01 \x01(\tH\x00\x12\x33\n\x0binline_data\x18\x02 \x01(\x0b\x32\x1c.executor_api_pb.DataPayloadH\x00\x42\x08\n\x06source"\xc2\x01\n\x08ReduceOp\x12\x0f\n\x02id\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x30\n\ncollection\x18\x02 \x03(\x0b\x32\x1c.executor_api_pb.FunctionArg\x12\x32\n\x07reducer\x18\x03 \x01(\x0b\x32\x1c.executor_api_pb.FunctionRefH\x01\x88\x01\x01\x12\x1a\n\rcall_metadata\x18\x04 \x01(\x0cH\x02\x88\x01\x01\x42\x05\n\x03_idB\n\n\x08_reducerB\x10\n\x0e_call_metadata"\x8b\x01\n\x14\x45xecutionPlanUpdates\x12\x35\n\x07updates\x18\x01 \x03(\x0b\x32$.executor_api_pb.ExecutionPlanUpdate\x12"\n\x15root_function_call_id\x18\x02 \x01(\tH\x00\x88\x01\x01\x42\x18\n\x16_root_function_call_id"\x8f\x05\n\x10\x41llocationResult\x12\x33\n\x08\x66unction\x18\x01 \x01(\x0b\x32\x1c.executor_api_pb.FunctionRefH\x01\x88\x01\x01\x12\x1a\n\rallocation_id\x18\x02 \x01(\tH\x02\x88\x01\x01\x12\x1d\n\x10\x66unction_call_id\x18\x03 \x01(\tH\x03\x88\x01\x01\x12\x17\n\nrequest_id\x18\x04 \x01(\tH\x04\x88\x01\x01\x12\x41\n\x0coutcome_code\x18\x05 \x01(\x0e\x32&.executor_api_pb.AllocationOutcomeCodeH\x05\x88\x01\x01\x12\x45\n\x0e\x66\x61ilure_reason\x18\x06 \x01(\x0e\x32(.executor_api_pb.AllocationFailureReasonH\x06\x88\x01\x01\x12-\n\x05value\x18\x07 \x01(\x0b\x32\x1c.executor_api_pb.DataPayloadH\x00\x12\x38\n\x07updates\x18\x08 \x01(\x0b\x32%.executor_api_pb.ExecutionPlanUpdatesH\x00\x12\x38\n\rrequest_error\x18\t \x01(\x0b\x32\x1c.executor_api_pb.DataPayloadH\x07\x88\x01\x01\x12"\n\x15\x65xecution_duration_ms\x18\n \x01(\x04H\x08\x88\x01\x01\x42\x0e\n\x0creturn_valueB\x0b\n\t_functionB\x10\n\x0e_allocation_idB\x13\n\x11_function_call_idB\r\n\x0b_request_idB\x0f\n\r_outcome_codeB\x11\n\x0f_failure_reasonB\x10\n\x0e_request_errorB\x18\n\x16_execution_duration_ms"\xab\x02\n\x13\x46unctionCallRequest\x12\x1e\n\x11parent_request_id\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x33\n\x08\x66unction\x18\x02 \x01(\x0b\x32\x1c.executor_api_pb.FunctionRefH\x01\x88\x01\x01\x12,\n\x06inputs\x18\x03 \x03(\x0b\x32\x1c.executor_api_pb.DataPayload\x12\x1a\n\rcall_metadata\x18\x04 \x01(\x0cH\x02\x88\x01\x01\x12$\n\x17source_function_call_id\x18\x05 \x01(\tH\x03\x88\x01\x01\x42\x14\n\x12_parent_request_idB\x0b\n\t_functionB\x10\n\x0e_call_metadataB\x1a\n\x18_source_function_call_id"c\n\x12\x46unctionCallResult\x12.\n\x06output\x18\x01 \x01(\x0b\x32\x1c.executor_api_pb.DataPayloadH\x00\x12\x13\n\texception\x18\x02 \x01(\tH\x00\x42\x08\n\x06result"h\n\x14\x46unctionCallResponse\x12\x10\n\x06update\x18\x01 \x01(\tH\x00\x12\x35\n\x06result\x18\x02 \x01(\x0b\x32#.executor_api_pb.FunctionCallResultH\x00\x42\x07\n\x05\x65vent*\xf0\x01\n\x13\x44\x61taPayloadEncoding\x12!\n\x1d\x44\x41TA_PAYLOAD_ENCODING_UNKNOWN\x10\x00\x12#\n\x1f\x44\x41TA_PAYLOAD_ENCODING_UTF8_JSON\x10\x01\x12#\n\x1f\x44\x41TA_PAYLOAD_ENCODING_UTF8_TEXT\x10\x02\x12\'\n#DATA_PAYLOAD_ENCODING_BINARY_PICKLE\x10\x03\x12$\n DATA_PAYLOAD_ENCODING_BINARY_ZIP\x10\x04\x12\x1d\n\x19\x44\x41TA_PAYLOAD_ENCODING_RAW\x10\x05*\xd6\x01\n\x08GPUModel\x12\x15\n\x11GPU_MODEL_UNKNOWN\x10\x00\x12\x1e\n\x1aGPU_MODEL_NVIDIA_A100_40GB\x10\x01\x12\x1e\n\x1aGPU_MODEL_NVIDIA_A100_80GB\x10\x02\x12\x1e\n\x1aGPU_MODEL_NVIDIA_H100_80GB\x10\x03\x12\x1d\n\x19GPU_MODEL_NVIDIA_TESLA_T4\x10\x04\x12\x1a\n\x16GPU_MODEL_NVIDIA_A6000\x10\x05\x12\x18\n\x14GPU_MODEL_NVIDIA_A10\x10\x06*\xb3\x01\n\x16\x46unctionExecutorStatus\x12$\n FUNCTION_EXECUTOR_STATUS_UNKNOWN\x10\x00\x12$\n FUNCTION_EXECUTOR_STATUS_PENDING\x10\x01\x12$\n FUNCTION_EXECUTOR_STATUS_RUNNING\x10\x02\x12\'\n#FUNCTION_EXECUTOR_STATUS_TERMINATED\x10\x03*\x94\x04\n!FunctionExecutorTerminationReason\x12\x30\n,FUNCTION_EXECUTOR_TERMINATION_REASON_UNKNOWN\x10\x00\x12\x46\nBFUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_INTERNAL_ERROR\x10\x01\x12\x46\nBFUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_FUNCTION_ERROR\x10\x02\x12H\nDFUNCTION_EXECUTOR_TERMINATION_REASON_STARTUP_FAILED_FUNCTION_TIMEOUT\x10\x03\x12\x32\n.FUNCTION_EXECUTOR_TERMINATION_REASON_UNHEALTHY\x10\x0c\x12\x37\n3FUNCTION_EXECUTOR_TERMINATION_REASON_INTERNAL_ERROR\x10\r\x12\x39\n5FUNCTION_EXECUTOR_TERMINATION_REASON_FUNCTION_TIMEOUT\x10\x0e\x12;\n7FUNCTION_EXECUTOR_TERMINATION_REASON_FUNCTION_CANCELLED\x10\x0f*\xa5\x01\n\x0e\x45xecutorStatus\x12\x1b\n\x17\x45XECUTOR_STATUS_UNKNOWN\x10\x00\x12\x1f\n\x1b\x45XECUTOR_STATUS_STARTING_UP\x10\x01\x12\x1b\n\x17\x45XECUTOR_STATUS_RUNNING\x10\x02\x12\x1b\n\x17\x45XECUTOR_STATUS_DRAINED\x10\x03\x12\x1b\n\x17\x45XECUTOR_STATUS_STOPPED\x10\x04*\x86\x01\n\x15\x41llocationOutcomeCode\x12#\n\x1f\x41LLOCATION_OUTCOME_CODE_UNKNOWN\x10\x00\x12#\n\x1f\x41LLOCATION_OUTCOME_CODE_SUCCESS\x10\x01\x12#\n\x1f\x41LLOCATION_OUTCOME_CODE_FAILURE\x10\x02*\xe9\x02\n\x17\x41llocationFailureReason\x12%\n!ALLOCATION_FAILURE_REASON_UNKNOWN\x10\x00\x12,\n(ALLOCATION_FAILURE_REASON_INTERNAL_ERROR\x10\x01\x12,\n(ALLOCATION_FAILURE_REASON_FUNCTION_ERROR\x10\x02\x12.\n*ALLOCATION_FAILURE_REASON_FUNCTION_TIMEOUT\x10\x03\x12+\n\'ALLOCATION_FAILURE_REASON_REQUEST_ERROR\x10\x04\x12\x32\n.ALLOCATION_FAILURE_REASON_ALLOCATION_CANCELLED\x10\x05\x12:\n6ALLOCATION_FAILURE_REASON_FUNCTION_EXECUTOR_TERMINATED\x10\x06\x32\xe3\x02\n\x0b\x45xecutorAPI\x12t\n\x15report_executor_state\x12+.executor_api_pb.ReportExecutorStateRequest\x1a,.executor_api_pb.ReportExecutorStateResponse"\x00\x12z\n\x1bget_desired_executor_states\x12\x30.executor_api_pb.GetDesiredExecutorStatesRequest\x1a%.executor_api_pb.DesiredExecutorState"\x00\x30\x01\x12\x62\n\x0finvoke_function\x12$.executor_api_pb.FunctionCallRequest\x1a%.executor_api_pb.FunctionCallResponse"\x00\x30\x01\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "indexify.proto.executor_api_pb2", _globals
)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_EXECUTORSTATE_LABELSENTRY"]._loaded_options = None
    _globals["_EXECUTORSTATE_LABELSENTRY"]._serialized_options = b"8\001"
    _globals["_DATAPAYLOADENCODING"]._serialized_start = 6138
    _globals["_DATAPAYLOADENCODING"]._serialized_end = 6378
    _globals["_GPUMODEL"]._serialized_start = 6381
    _globals["_GPUMODEL"]._serialized_end = 6595
    _globals["_FUNCTIONEXECUTORSTATUS"]._serialized_start = 6598
    _globals["_FUNCTIONEXECUTORSTATUS"]._serialized_end = 6777
    _globals["_FUNCTIONEXECUTORTERMINATIONREASON"]._serialized_start = 6780
    _globals["_FUNCTIONEXECUTORTERMINATIONREASON"]._serialized_end = 7312
    _globals["_EXECUTORSTATUS"]._serialized_start = 7315
    _globals["_EXECUTORSTATUS"]._serialized_end = 7480
    _globals["_ALLOCATIONOUTCOMECODE"]._serialized_start = 7483
    _globals["_ALLOCATIONOUTCOMECODE"]._serialized_end = 7617
    _globals["_ALLOCATIONFAILUREREASON"]._serialized_start = 7620
    _globals["_ALLOCATIONFAILUREREASON"]._serialized_end = 7981
    _globals["_DATAPAYLOAD"]._serialized_start = 55
    _globals["_DATAPAYLOAD"]._serialized_end = 502
    _globals["_GPURESOURCES"]._serialized_start = 504
    _globals["_GPURESOURCES"]._serialized_end = 605
    _globals["_HOSTRESOURCES"]._serialized_start = 608
    _globals["_HOSTRESOURCES"]._serialized_end = 802
    _globals["_ALLOWEDFUNCTION"]._serialized_start = 805
    _globals["_ALLOWEDFUNCTION"]._serialized_end = 1016
    _globals["_FUNCTIONEXECUTORRESOURCES"]._serialized_start = 1019
    _globals["_FUNCTIONEXECUTORRESOURCES"]._serialized_end = 1235
    _globals["_FUNCTIONREF"]._serialized_start = 1238
    _globals["_FUNCTIONREF"]._serialized_end = 1445
    _globals["_FUNCTIONEXECUTORDESCRIPTION"]._serialized_start = 1448
    _globals["_FUNCTIONEXECUTORDESCRIPTION"]._serialized_end = 1995
    _globals["_FUNCTIONEXECUTORSTATE"]._serialized_start = 1998
    _globals["_FUNCTIONEXECUTORSTATE"]._serialized_end = 2333
    _globals["_EXECUTORSTATE"]._serialized_start = 2336
    _globals["_EXECUTORSTATE"]._serialized_end = 3054
    _globals["_EXECUTORSTATE_LABELSENTRY"]._serialized_start = 2867
    _globals["_EXECUTORSTATE_LABELSENTRY"]._serialized_end = 2912
    _globals["_EXECUTORUPDATE"]._serialized_start = 3056
    _globals["_EXECUTORUPDATE"]._serialized_end = 3177
    _globals["_REPORTEXECUTORSTATEREQUEST"]._serialized_start = 3180
    _globals["_REPORTEXECUTORSTATEREQUEST"]._serialized_end = 3371
    _globals["_REPORTEXECUTORSTATERESPONSE"]._serialized_start = 3373
    _globals["_REPORTEXECUTORSTATERESPONSE"]._serialized_end = 3402
    _globals["_ALLOCATION"]._serialized_start = 3405
    _globals["_ALLOCATION"]._serialized_end = 3943
    _globals["_GETDESIREDEXECUTORSTATESREQUEST"]._serialized_start = 3945
    _globals["_GETDESIREDEXECUTORSTATESREQUEST"]._serialized_end = 4020
    _globals["_DESIREDEXECUTORSTATE"]._serialized_start = 4023
    _globals["_DESIREDEXECUTORSTATE"]._serialized_end = 4199
    _globals["_EXECUTIONPLANUPDATE"]._serialized_start = 4202
    _globals["_EXECUTIONPLANUPDATE"]._serialized_end = 4330
    _globals["_FUNCTIONCALL"]._serialized_start = 4333
    _globals["_FUNCTIONCALL"]._serialized_end = 4523
    _globals["_FUNCTIONARG"]._serialized_start = 4525
    _globals["_FUNCTIONARG"]._serialized_end = 4629
    _globals["_REDUCEOP"]._serialized_start = 4632
    _globals["_REDUCEOP"]._serialized_end = 4826
    _globals["_EXECUTIONPLANUPDATES"]._serialized_start = 4829
    _globals["_EXECUTIONPLANUPDATES"]._serialized_end = 4968
    _globals["_ALLOCATIONRESULT"]._serialized_start = 4971
    _globals["_ALLOCATIONRESULT"]._serialized_end = 5626
    _globals["_FUNCTIONCALLREQUEST"]._serialized_start = 5629
    _globals["_FUNCTIONCALLREQUEST"]._serialized_end = 5928
    _globals["_FUNCTIONCALLRESULT"]._serialized_start = 5930
    _globals["_FUNCTIONCALLRESULT"]._serialized_end = 6029
    _globals["_FUNCTIONCALLRESPONSE"]._serialized_start = 6031
    _globals["_FUNCTIONCALLRESPONSE"]._serialized_end = 6135
    _globals["_EXECUTORAPI"]._serialized_start = 7984
    _globals["_EXECUTORAPI"]._serialized_end = 8339
# @@protoc_insertion_point(module_scope)
